@@ -45,6 +45,9 @@ XmmsKde::XmmsKde(const QString& configFile, Type type, int actions,
   stopPressed  = false;
   loopedOnce   = false;
 
+  aboutData = NULL;
+  playList = NULL;
+
   seekTime = 5000;
     
   setAcceptDrops(true);
@@ -98,7 +101,7 @@ XmmsKde::XmmsKde(const QString& configFile, Type type, int actions,
    * search for themes in the home directory
    */
   themeDir = QDir(homeDir + "/.xmms-kde/", "*.rc");
-
+  
   if (themeDir.exists()) {
     for (unsigned int k = 0; k < themeDir.count(); k++) {
       themeMenu->insertItem(themeDir[k], i + k);
@@ -114,7 +117,7 @@ XmmsKde::XmmsKde(const QString& configFile, Type type, int actions,
   titleDefaultScrollSpeed= mainConfig->readNumEntry("scrollspeed", 1);
   displayElapsedTime = mainConfig->readNumEntry("displayelapsedtime", 1);
   
-  title = "";
+  title = "xmms-kde";
 
   mainConfig->setGroup("PLAYER");
   PlayerInterface::Player currentPlayer = 
@@ -168,7 +171,7 @@ XmmsKde::XmmsKde(const QString& configFile, Type type, int actions,
 		      CTRL+SHIFT+Key_S,
 		      (QObject *) this, 
 		      SLOT(next()));
-
+  
   globalAccel->insert("xmms-kde back", i18n("previous song"),
 		      i18n("Jump to previous song in playlist"),
 		      CTRL+SHIFT+Key_A,
@@ -217,6 +220,13 @@ XmmsKde::XmmsKde(const QString& configFile, Type type, int actions,
 		      CTRL+SHIFT+Key_R, 
 		      (QObject *) this, 
 		      SLOT(playpause()));
+
+  globalAccel->insert("xmms-kde jump to file", i18n("jump to file"),
+		      i18n("Jump to a specific file in the playlist"),
+		      CTRL+SHIFT+Key_J, 
+		      CTRL+SHIFT+Key_J, 
+		      (QObject *) this, 
+		      SLOT(jumpToFile()));
   
   globalAccel->insert("xmms-kde volume plus", i18n("increase volume"), 
 		      i18n("Increase the volume"),
@@ -231,6 +241,13 @@ XmmsKde::XmmsKde(const QString& configFile, Type type, int actions,
 		      CTRL+SHIFT+Key_Y,
 		      (QObject *) this, 
 		      SLOT(volm()));
+		      
+  globalAccel->insert("xmms-kde show track", i18n("Show On-Screen-Display"),
+  		      i18n("Show the current track on the display"),
+		      CTRL+SHIFT+Key_T,
+		      CTRL+SHIFT+Key_T,
+		      (QObject *) this,
+		      SLOT(showOSD()));
 
   globalAccel->insert("xmms-kde popup database query", i18n("show database window"), 
 		      i18n("Show the database window"),
@@ -246,6 +263,27 @@ XmmsKde::XmmsKde(const QString& configFile, Type type, int actions,
 		      (QObject *) db, 
 		      SLOT(updateDatabase()));
 
+  globalAccel->insert("xmms-kde toggle shuffle", i18n("toggle shuffle"), 
+		      i18n("Toggle Shuffle"),
+		      CTRL+SHIFT+Key_F,
+		      CTRL+SHIFT+Key_F,
+		      (QObject *) this, 
+		      SLOT(toggleShuffle()));
+
+  globalAccel->insert("xmms-kde toggle repeat", i18n("toggle repeat"), 
+		      i18n("Toggle Repeat"),
+		      CTRL+SHIFT+Key_G,
+		      CTRL+SHIFT+Key_G,
+		      (QObject *) this, 
+		      SLOT(toggleRepeat()));
+
+  globalAccel->insert("xmms-kde show lyrics", i18n("show lyrics"), 
+		      i18n("Show Lyrics"),
+		      CTRL+SHIFT+Key_L,
+		      CTRL+SHIFT+Key_L,
+		      (QObject *) this, 
+		      SLOT(showLyrics()));
+
   globalAccel->readSettings(mainConfig);
 
   globalAccel->updateConnections();
@@ -257,10 +295,13 @@ XmmsKde::XmmsKde(const QString& configFile, Type type, int actions,
   } else {
     globalAccel->setEnabled(false);
   }
-
+  
   mainConfig->setGroup("XMMS");
   hideXmms = mainConfig->readBoolEntry("hide", false);
 
+  mainConfig->setGroup("MPLAYER");
+  mplayerBorderless = mainConfig->readBoolEntry("noborder", false);
+  mplayerScreen = mainConfig->readBoolEntry("showscreen", false);
 
   player = NULL;
 
@@ -280,12 +321,20 @@ XmmsKde::XmmsKde(const QString& configFile, Type type, int actions,
 			     playListIcon);
 #endif
 
+#ifdef HAVE_AMAROK
+  if (currentPlayer == PlayerInterface::AMAROK) player = new AmarokPlayer();
+#endif
+
+#ifdef HAVE_MPLAYER
+  if (currentPlayer == PlayerInterface::MPLAYER) player = new MPlayer();
+#endif
+
 #ifdef HAVE_NOATUN
   if (!player) player = new NoatunPlayer();
 #endif
 
   if (!player) player = new PlayerInterface();
-
+  
   qDebug("xmms-kde: player: %d", player->is());
 
   osd = new OSDFrame(mainConfig);
@@ -295,13 +344,20 @@ XmmsKde::XmmsKde(const QString& configFile, Type type, int actions,
   if (query->getPopup() == 0) // kicker startup
     query->popup();
 
-  title = QString("");
+  title = "xmms-kde";
 
-  aboutData = new KAboutData("XmmsKde", "XmmsKde", 
-			     "3.1", "MP3 Player Applet",
-			     KAboutData::License_GPL_V2, "(c) 2002, Flo Niebling", 0, "http://xmms-kde.sourceforge.net", "tranqlzer@users.sourceforge.net");
-  
-  aboutData->addAuthor("Flo Niebling", 0, "tranqlzer@users.sourceforge.net");  
+  if (!aboutData) {
+    aboutData = new KAboutData("XmmsKde", "XmmsKde", 
+			       "3.1", "MP3 Player Applet",
+			       KAboutData::License_GPL_V2, "(c) 2002, Flo Niebling", 0, "http://xmms-kde.sourceforge.net", "tranqlzer@users.sourceforge.net");
+    
+    aboutData->addAuthor("Flo Niebling", 0, "tranqlzer@users.sourceforge.net");
+    aboutData->addAuthor("Rajendra R Patil", "mplayer support",
+			 "rajendra_p123@yahoo.com");
+
+  }
+
+  lyrics = new SongLyrics(mainConfig);
 }
 
 
@@ -445,7 +501,16 @@ void XmmsKde::preferences() {
 #else
       qDebug("xmms-kde: don't have smpeg");
 #endif
-      if (!player) player = new NoatunPlayer();
+
+#ifdef HAVE_AMAROK
+      if (newPlayer == PlayerInterface::AMAROK) player = new AmarokPlayer();
+#endif
+
+#ifdef HAVE_MPLAYER
+      if (newPlayer == PlayerInterface::MPLAYER) player = new MPlayer();
+#endif
+
+      if (!player) player = new PlayerInterface();
 
       if (query)
 	query->setPlayer(player);
@@ -524,11 +589,11 @@ void XmmsKde::paintEvent(QPaintEvent*) {
 void XmmsKde::paint() {
   
   QPixmap pmt(theme.myWidth, theme.myHeight);
-  pmt.fill(this, theme.myWidth, theme.myHeight);
+  pmt.fill(this, 0, 0);
 
   QPainter pt;
   pt.begin(&pmt);
-  pt.setClipRect(theme.titleClipRect);
+
   pt.setClipping(false);
 
   /*
@@ -567,17 +632,18 @@ void XmmsKde::paint() {
 		    *theme.shuffleOnPixmap );
     }
     
-    
     /*
      * draw the title
      */
     pt.setPen(theme.titleColor);
     pt.setFont(theme.titleFont);
+    fm = pt.fontMetrics();
     QFontInfo fi = pt.fontInfo();
     if (theme.titleYPos > -1) {
       pt.setClipping(true);
+      pt.setClipRect(theme.titleClipRect);
       titleWidth = fm.boundingRect(title).width();
-      pt.drawText(titleXPosition, theme.titleYPos + fi.pointSize(), title);
+      pt.drawText(titleXPosition, theme.titleYPos + fm.ascent(), title);
       pt.setClipping(false);
     }
     
@@ -610,13 +676,14 @@ void XmmsKde::paint() {
       pt.setPen(theme.timeColor); // set the selected pen.
       pt.setFont(theme.timeFont);
       fi = pt.fontInfo();
+      fm = pt.fontMetrics();
       if (displayElapsedTime) {
 	char pTime[5];
 	sprintf(pTime, "%02d:%02d", 
 		(int) (currentTime / 60000),
 		(int) (currentTime / 1000) % 60);
 	
-	pt.drawText(theme.timePlayedPos.x(), theme.timePlayedPos.y() + fi.pointSize(), pTime);
+	pt.drawText(theme.timePlayedPos.x(), theme.timePlayedPos.y() + fm.ascent(), pTime);
       } else {
 	int remainingTime = trackTime - currentTime;
 	char pTime[6];
@@ -624,7 +691,7 @@ void XmmsKde::paint() {
 		(int) (remainingTime / 60000),
 		(int) (remainingTime / 1000) % 60);
 	
-	pt.drawText(theme.timePlayedPos.x(), theme.timePlayedPos.y() + fi.pointSize(), pTime);
+	pt.drawText(theme.timePlayedPos.x(), theme.timePlayedPos.y() + fm.ascent(), pTime);
       }
     }
 
@@ -637,7 +704,7 @@ void XmmsKde::paint() {
 	      (int) (trackTime / 60000),
 	      (int) (trackTime / 1000) % 60);
       
-      pt.drawText( theme.timeTotalPos.x(), theme.timeTotalPos.y() + fi.pointSize(), totalTime);
+      pt.drawText( theme.timeTotalPos.x(), theme.timeTotalPos.y() + fm.ascent(), totalTime);
     }      
 
   pt.end();
@@ -693,13 +760,13 @@ void XmmsKde::scroll() {
 	/*
 	 * sinus-scrolling
 	 */
-	int scrolllength = titleWidth - theme.myWidth + 30;
+	int scrolllength = titleWidth - theme.titleClipRect.width() + 30;
 	
 	titleSinArg += M_PI / 3 * ((double) titleScrollSpeed) / scrolllength;
 	if (titleSinArg > 2 * M_PI) titleSinArg = 0;
 	
-	titleXPosition = (int) rint(15 - scrolllength * 
-				    (sin(titleSinArg) * sin(titleSinArg)));
+	titleXPosition = theme.titleClipRect.left() + 
+	  (int) rint(15 - scrolllength * (sin(titleSinArg)*sin(titleSinArg)));
       } else if (titleScrollMode == 3) {
         /*
 	 * loop once scrolling
@@ -744,13 +811,22 @@ void XmmsKde::receive() {
     newTitle = title;
   }
 
-  // if title has changed, display osd
-  if (newTitle && title.compare(newTitle) && title.compare("xmms-kde")) {
+  // if title has changed, display osd and lyrics
+  if (newTitle && title.compare(newTitle) && newTitle.compare("xmms-kde")) {
     if (osd) {
       osd->startSequence(newTitle);
       titleXPosition     = theme.titleClipRect.right();
       loopedOnce = false;
     }
+
+    if (lyrics && lyrics->isVisible()) {
+      QString lartist = newTitle.section(" - ", 0, 0);
+      QString ltitle = newTitle.section(" - ", 1, 10);
+      
+      qDebug("search: [%s] [%s]", lartist.latin1(), ltitle.latin1());
+
+      lyrics->request(lartist, ltitle);
+    }
   }
 
   title = newTitle;
@@ -764,6 +840,20 @@ void XmmsKde::receive() {
 }
 
 
+/*
+ * Show the on-screen-display
+ *
+ */
+void XmmsKde::showOSD( ) {
+
+  if (!osd) 
+    return;
+  
+  osd->startSequence(title);
+  titleXPosition = theme.titleClipRect.right();
+  loopedOnce = false;
+}
+
 
 /*
  * mouse handling
@@ -903,6 +993,7 @@ void XmmsKde::mousePressEvent(QMouseEvent *event) {
 
     QPopupMenu *menu = new QPopupMenu(this);
 
+    menu->insertItem(i18n("Jump to file"), this, SLOT(jumpToFile()));
     menu->insertItem(i18n("Play"), this, SLOT(play()));
     menu->insertItem(i18n("Previous"), this, SLOT(previous()));
     menu->insertItem(i18n("Next"), this, SLOT(next()));
@@ -911,8 +1002,15 @@ void XmmsKde::mousePressEvent(QMouseEvent *event) {
     menu->insertItem(i18n("Exit Player"), this, SLOT(exitPlayer()));
     menu->insertSeparator();
 
-    menu->insertItem(i18n("Database Query Window"), this, SLOT(showDB()));
-    menu->insertItem(i18n("Update Database"), db, SLOT(updateDatabase()));
+    if (db) {
+      menu->insertItem(i18n("Database Query Window"), this, SLOT(showDB()));
+      menu->insertItem(i18n("Update Database"), db, SLOT(updateDatabase()));
+    }
+    
+    if (lyrics) {
+      menu->insertItem(i18n("Show Lyrics"), this, SLOT(showLyrics()));
+    }
+
     menu->insertSeparator();
 
     menu->insertItem(i18n("Help"), this, SLOT(help()));
@@ -1039,6 +1137,9 @@ void XmmsKde::saveConfig() {
     if (query)
       query->writeConfig();
 
+    if (lyrics)
+      lyrics->writeConfig();
+
     if (globalAccel) {
       mainConfig->setGroup("ACCEL");
       mainConfig->writeEntry("enable", globalAccel->isEnabled());
@@ -1063,6 +1164,10 @@ void XmmsKde::saveConfig() {
     }
 #endif
     
+    mainConfig->setGroup("MPLAYER");
+    mainConfig->writeEntry("noborder", mplayerBorderless);
+    mainConfig->writeEntry("showscreen", mplayerScreen);
+ 
     mainConfig->setGroup("Global Keys");
 
     globalAccel->updateConnections();    
@@ -1092,6 +1197,9 @@ void XmmsKde::dragEnterEvent(QDragEnterEvent* event) {
  */
 void XmmsKde::dropEvent(QDropEvent* event) {
 
+  if (!player)
+    return;
+
   QString text;
 
   if (QTextDrag::decode(event, text)) {  
@@ -1118,6 +1226,14 @@ void XmmsKde::dropEvent(QDropEvent* event) {
   }
 }
 
+void XmmsKde::jumpto(int pos) {
+
+  player->jumpto(pos);
+
+   if (!player->isPlaying())
+    player->play();  
+}
+
 void XmmsKde::previous() {
   player->previous();
 }
@@ -1203,18 +1319,220 @@ void XmmsKde::toggleRepeat() {
  * load a theme from a file
  *
  */
-void XmmsKde::loadTheme(QString themeName)
-{
-    setTimers(false);
-    theme.loadTheme(themeName, dataDirs);
-    
-    setTimers(true);
-    QSize size(theme.myWidth, theme.myHeight);
-    resize(size);
+void XmmsKde::loadTheme(QString themeName) {
+
+  setTimers(false);
+  theme.loadTheme(themeName, dataDirs);
+  
+  setTimers(true);
+  QSize size(theme.myWidth, theme.myHeight);
+  resize(size);
 }
 
 
 
+void XmmsKde::jumpToFile() {
+
+  if (!player)
+    return;
+
+  if (playList)
+    delete playList;
+
+  playList = new PlayList();
+  QStringList *list = new QStringList();
+
+  playList->resize(400, 300);
+
+  QRect rect = QApplication::desktop()->geometry();
+  playList->move(rect.center() - playList->rect().center());
+
+  int length = player->getPlaylistLength();
+
+  for (int index = 0; index < length; index ++)
+    list->append(player->getPlaylistTitle(index));
+
+  playList->setContents(list);
+  playList->setSelected(player->getPlaylistPos(), true);
+  playList->centerCurrentItem();
+
+  connect((QObject *) playList, SIGNAL(sigSelectionChanged(int)),
+	  (QObject *) this, SLOT(jumpto(int)));
+
+  KWin::setOnDesktop(playList->winId(), KWin::currentDesktop());
+
+  playList->show();
+}
+
+void XmmsKde::showLyrics() {
+
+  if (lyrics) {
+    if (!lyrics->isVisible()) {
+      QString lartist = title.section(" - ", 0, 0);
+      QString ltitle = title.section(" - ", 1, 10);
+    
+      qDebug("search: [%s] [%s]\n", lartist.latin1(), ltitle.latin1());
+    
+      lyrics->request(lartist, ltitle);
+    }
+
+    lyrics->show();
+  }
+}
+
+
+PlayList::PlayList() 
+  : KDialogBase (0, 0, false, i18n("Jump to file"), Ok | Cancel, Ok) {
+
+  contents = 0;
+  setFocusProxy(0);
+
+  QVBox *box = new QVBox(this);
+  
+  line = new QLineEdit(box);
+  playListBox = new QListBox(box);
+
+  setMainWidget(box);
+
+  connect((QObject *) playListBox, SIGNAL(returnPressed(QListBoxItem *)),
+	  (QObject *) this, SLOT(selectionChanged()));
+
+  connect((QObject *) line, SIGNAL(returnPressed()),
+	  (QObject *) this, SLOT(selectionChanged()));
+
+  connect((QObject *) line, SIGNAL(textChanged(const QString &)),
+	  (QObject *) this, SLOT(textChanged(const QString &)));
+
+  connect((QObject *) playListBox, SIGNAL(doubleClicked(QListBoxItem *)),
+	  (QObject *) this, SLOT(selectionChanged()));
+
+  connect((QObject *) this, SIGNAL(okClicked()),
+	  (QObject *) this, SLOT(selectionChanged()));
+
+  connect((QObject *) this, SIGNAL(cancelClicked()),
+	  (QObject *) this, SLOT(cancel()));
+
+
+  setCaption(i18n("Jump to file"));
+  line->setFocus();
+
+  //KWin::setIcons(winId(), *icon, *icon);
+}
+
+PlayList::~PlayList() {
+}
+
+void PlayList::setContents(QStringList *list) {
+
+  playListBox->clear();
+
+  playListBox->insertStringList(*list);
+
+  if (contents)
+    delete contents;
+
+  contents = list;
+}
+
+int PlayList::getSelected() {
+  return playListBox->currentItem();
+}
+
+int PlayList::setSelected(int pos, bool status) {
+  playListBox->setSelected(pos, status);
+  return pos;
+}
+
+void PlayList::clear() {
+  playListBox->clear();
+}
+
+
+void PlayList::centerCurrentItem() {
+  playListBox->centerCurrentItem();
+}
+
+void PlayList::selectionChanged() {
+  
+  // find the currently selected item in our contents list
+  QString current(playListBox->currentText());
+  int index = 0;
+
+  for (QStringList::Iterator it = contents->begin(); it != contents->end(); ++it ) {
+    
+    if (!current.compare(*it))
+      break;
+    
+    index ++;
+  }
+      
+  emit sigSelectionChanged(index);
+  hide();
+}
+
+void PlayList::textChanged(const QString & s) {
+
+  QStringList::Iterator it;
+
+  QStringList exlist = QStringList::split(" ", s);
+
+  /*
+  if (exlist.size() == 0)
+    return;
+  */
+
+    // build a number of regular expressions from the input string
+  QRegExp **e = new QRegExp*[exlist.size()];
+
+  int index = 0;
+  for (it = exlist.begin(); it != exlist.end(); ++it )
+    e[index++] = new QRegExp(*it, false);
+
+  playListBox->clear();
+
+  for (it = contents->begin(); it != contents->end(); ++it ) {
+
+    bool test = true;
+    for (int i = 0; i < index; i++)
+      if (e[i]->search(*it) < 0)
+	test = false;
+    
+    if (test)
+      playListBox->insertItem(*it);
+  }
+
+  delete[] e;
+}
+
+void PlayList::keyPressEvent(QKeyEvent *e) {
+
+  if (e->key() == Qt::Key_Escape) {
+    e->accept();
+    hide();
+  } else if (e->key() == Qt::Key_Up) {
+    int s = playListBox->currentItem();
+    playListBox->setFocus();
+    playListBox->setSelected(s - 1, true);
+  } else if (e->key() == Qt::Key_Down) {
+    int s = playListBox->currentItem();
+    playListBox->setFocus();
+    playListBox->setSelected(s + 1, true);
+    e->ignore();
+  } else {
+    e->ignore();
+  }
+}
+
+void PlayList::cancel() {
+
+  hide();
+}
+
+void PlayList::focusOutEvent(QFocusEvent *e) {
+
+  hide();
+}
+
 /********************* THEME CLASS  *************************/
 XmmsKdeTheme::XmmsKdeTheme() {
   mainFont = QFont("helvetica", 10);
