@@ -19,8 +19,10 @@
 /*
  * TODO: - still (rare) crashes when updating db and dialog is closed (???)
  *       - single inserts fail when querying at the same time (???)
- *         happens also when queried from command line so this seems to be 
+ *         happens also when queried from command line so this seems to be
  *         an sqlite bug
+ *       - database locking on nfs doesn't work (sqlite/nfs bug)
+ *       - failure to insert non-ascii characters in pathname (locale problem?)
  */
 
 #include <config.h>
@@ -33,11 +35,23 @@
 #include <qradiobutton.h>
 #include <qregexp.h>
 #include <qhbuttongroup.h>
-#include <qdatetime.h>
 #include <qpixmap.h>
+#include <qlineedit.h>
+#include <qtimer.h>
 
 #include <kfilemetainfo.h>
 #include <kwin.h>
+#include <kapplication.h>
+
+#ifdef HAVE_ID3LIB
+#include <id3/tag.h>
+#endif
+
+#ifdef HAVE_VORBIS
+#include <ogg/ogg.h>
+#include <vorbis/codec.h>
+#include <vorbis/vorbisfile.h>
+#endif
 
 #include <configdialog.h>
 
@@ -62,7 +76,7 @@ static int callback(void *NotUsed, int argc, char **argv, char **azColName) {
 
 /**
  * the table in the database has to be of the following format:
- * 
+ *
  * create table music
  * ( filename VARCHAR(255),
  *   title VARCHAR(255),
@@ -77,6 +91,8 @@ static int callback(void *NotUsed, int argc, char **argv, char **azColName) {
 XmmsKdeDB::XmmsKdeDB(KConfig *conf, QPixmap *ico) {
 
   db = NULL;
+  querydb = NULL;
+  
   pathListBox = NULL;
   insertThread = NULL;
   statusFrame = NULL;
@@ -98,7 +114,7 @@ XmmsKdeDB::~XmmsKdeDB() {
 }
 
 bool XmmsKdeDB::connectDB() {
-  
+
   char *error = 0;
 
   if (enable) {
@@ -106,7 +122,7 @@ bool XmmsKdeDB::connectDB() {
     if (connected) {
       disconnectDB();
     }
-    
+
     qDebug("xmms-kde: trying to open database");
     db = sqlite_open(name, 0, &error);
 
@@ -115,14 +131,13 @@ bool XmmsKdeDB::connectDB() {
       error = 0;
     }
 
-    qDebug("xmms-kde: creating table");
-    sqlite_exec(db, "create table music (filename string, title string, artist string, album string, track string, year string, genre string, comment string)", 0, 0, &error);
+    sqlite_exec(db, "CREATE TABLE music (filename STRING UNIQUE ON CONFLICT REPLACE, title STRING, artist STRING, album STRING, track STRING, year STRING, genre STRING, comment STRING)", 0, 0, &error);
 
     if (error) {
       free(error);
       error = 0;
     }
-    
+
     qDebug("xmms-kde: table created");
 
     querydb = sqlite_open(name, 0, &error);
@@ -146,7 +161,7 @@ bool XmmsKdeDB::connectDB() {
 }
 
 void XmmsKdeDB::disconnectDB() {
-  
+
   if (enable) {
     if (connected) {
       connected = false;
@@ -167,8 +182,8 @@ bool XmmsKdeDB::isConnectedDB() {
 
 QWidget* XmmsKdeDB::getConfigurationWidget(QWidget *parent) {
 
-  QWidget *dbWidget = new QWidget(parent);  
-  
+  QWidget *dbWidget = new QWidget(parent);
+
   QVBoxLayout *vbox = new QVBoxLayout(dbWidget, 5);
 
   QHBoxLayout *box = new QHBoxLayout(vbox, 10);
@@ -177,12 +192,7 @@ QWidget* XmmsKdeDB::getConfigurationWidget(QWidget *parent) {
 
   box->addWidget(enableBox);
   box->addStretch(10);
-  /*
-  QPushButton *dbUpdateButton = new QPushButton(i18n("Update Database"), dbWidget);
-  connect(dbUpdateButton, SIGNAL(clicked()), this, SLOT(updateDatabase()));
 
-  box->addWidget(dbUpdateButton);
-  */
   connect(enableBox, SIGNAL(clicked()), this, SLOT(configurationChanged()));
 
   pathListBox = new QListBox(dbWidget);
@@ -206,13 +216,13 @@ QWidget* XmmsKdeDB::getConfigurationWidget(QWidget *parent) {
 
 void XmmsKdeDB::configurationChanged() {
 
-  enable = enableBox->isChecked();  
+  enable = enableBox->isChecked();
 }
 
 void XmmsKdeDB::updateDatabase() {
 
   if (insertThread != NULL) {
-    
+
     if (insertThread->running()) {
       if (statusFrame)
 	statusFrame->show();
@@ -233,7 +243,7 @@ void XmmsKdeDB::updateDatabase() {
     delete statusFrame;
     statusFrame = NULL;
   }
-  
+
   if (pathList.empty())
     return;
 
@@ -241,18 +251,18 @@ void XmmsKdeDB::updateDatabase() {
   statusFrame->setCaption(i18n("Database status"));
   QVBoxLayout *box = new QVBoxLayout(statusFrame, 10);
   QHBoxLayout *hbox = new QHBoxLayout(box, 20);
-  
+
   QLabel *pixLabel = new QLabel(statusFrame);
   pixLabel->setPixmap(*icon);
-  
+
   QVBoxLayout *vbox2 = new QVBoxLayout(hbox, 0);
   QVBoxLayout *vbox3 = new QVBoxLayout(hbox, 0);
-  
+
   QLabel *label = new QLabel(i18n("Updating database (this will take a while)."), statusFrame);
   StatusLabel *statusLabel = new StatusLabel("", statusFrame, 45);
-  
+
   StatusLabel *fileLabel = new StatusLabel("", statusFrame, 45);
-  
+
   vbox2->addWidget(pixLabel);
   vbox3->addWidget(label);
   vbox3->addWidget(statusLabel);
@@ -260,32 +270,35 @@ void XmmsKdeDB::updateDatabase() {
 
   ProgressLabel *progressLabel = new ProgressLabel(statusFrame);
   box->addWidget(progressLabel);
-  
+
   QHBoxLayout *hbox2 = new QHBoxLayout(box, 10);
-  
+
   QPushButton *okButton = new QPushButton(i18n(i18n("OK")), statusFrame);
   okButton->setDefault(true);
   //QPushButton *cancelButton = new QPushButton(i18n("Cancel"), statusFrame);
-  
+
   hbox2->addStretch(10);
   hbox2->addWidget(okButton, 1);
   //hbox2->addWidget(cancelButton, 1);
-  
+
   connect(okButton, SIGNAL(clicked()), statusFrame, SLOT(hide()));
   //connect(cancelButton, SIGNAL(clicked()), this, SLOT(stopInsertThread()));
   //connect(cancelButton, SIGNAL(clicked()), statusFrame, SLOT(hide()));
-  
+
   statusFrame->show();
-  
+
   fileLabel->setMaximumSize(label->size());
   statusLabel->setMaximumSize(label->size());
   fileLabel->setMinimumSize(label->size());
   statusLabel->setMinimumSize(label->size());
-  
+
   insertThread = new InsertThread(db, pathList, statusLabel, fileLabel,
-				  progressLabel);
+				  progressLabel, updated);
   insertThread->start();
-  
+
+  updated = QDateTime::currentDateTime();
+
+  writeConfig();
   sync = true;
 }
 
@@ -294,9 +307,11 @@ void XmmsKdeDB::readConfig() {
   config->setGroup("DB");
   enable = config->readBoolEntry("enable", false);
   QString dbfile(locateLocal("data", "xmms-kde/music.db"));
+  //QString dbfile("/tmp/music.db");
   name = config->readEntry("name", dbfile);
   qDebug("xmms-kde: using database '" + dbfile + "'");
   pathList = config->readListEntry("paths");
+  updated = config->readDateTimeEntry("updated");
 }
 
 
@@ -304,8 +319,8 @@ void XmmsKdeDB::writeConfig() {
 
   config->setGroup("DB");
   config->writeEntry("enable", enable);
-
-  config->writeEntry("paths", pathList);  
+  config->writeEntry("paths", pathList);
+  config->writeEntry("updated", updated);
 }
 
 
@@ -339,46 +354,53 @@ void XmmsKdeDB::stopInsertThread() {
  // FIXME
   /*
   if (insertThread) {
-    
+
     if (insertThread->running())
       insertThread->terminate();
   }
   */
 }
 
+
 XmmsKdeDBQuery::XmmsKdeDBQuery(XmmsKdeDB *datab, PlayerInterface *p,
 			       QPixmap *icon, KConfig *conf)
-
-  : QVBox(0, "Database Query", WStyle_Title) {
+  : QSplitter(0, "Database Query") {
 
   config = conf;
   readConfig();
 
-  searchThread = NULL;
   db = datab;
   player = p;
 
-  resultBox = new QListBox(this);
+  QVBox *vbox1 = new QVBox(this);
+  vbox1->setSpacing(2);
+
+  firstCombo = new QComboBox(false, vbox1);
+  firstBox = new QListBox(vbox1);
+  firstBox->setSelectionMode(QListBox::Extended);
+  firstFilterEdit = new QLineEdit(vbox1);
+
+
+  QSplitter *vsplitter = new QSplitter(this);
+  vsplitter->setOrientation(QSplitter::Vertical);
+
+  QVBox *vbox2 = new QVBox(vsplitter);
+  vbox2->setSpacing(2);
+
+  secondCombo = new QComboBox(false, vbox2);
+  secondBox = new QListBox(vbox2);
+  secondBox->setSelectionMode(QListBox::Extended);
+  secondFilterEdit = new QLineEdit(vbox2);
+
+  QVBox *vbox3 = new QVBox(vsplitter);
+  resultBox = new QListBox(vbox3);
   resultBox->setSelectionMode(QListBox::Extended);
+  resultFilterEdit = new QLineEdit(vbox3);
 
-  QHBox *hbox = new QHBox(this);
-  hbox->setSpacing(0);
+  QHBox *hbox = new QHBox(vbox3);
 
-  buttons = new QHButtonGroup(this);
-  buttons->setExclusive(true);
-
-  QRadioButton *titleButton = new QRadioButton(i18n("Title"), buttons, "title");
-  QRadioButton *artistButton = new QRadioButton(i18n("Artist"), buttons, "artist");
-  QRadioButton *albumButton = new QRadioButton(i18n("Album"), buttons, "album");
-  QRadioButton *genreButton = new QRadioButton(i18n("Genre"), buttons, "genre");
-
-  buttons->setButton(0);
-
-  QHBox *box = new QHBox(this);
-  query = new QLineEdit("", box);
-
-  QPushButton *set = new QPushButton(i18n("set"), box);
-  QPushButton *add = new QPushButton(i18n("add"), box);
+  QPushButton *set = new QPushButton(i18n("set"), hbox);
+  QPushButton *add = new QPushButton(i18n("add"), hbox);
 
   connect((QObject *) set, SIGNAL(clicked()),
 	  this, SLOT(setPlayList()));
@@ -386,19 +408,55 @@ XmmsKdeDBQuery::XmmsKdeDBQuery(XmmsKdeDB *datab, PlayerInterface *p,
   connect((QObject *) add, SIGNAL(clicked()),
 	  this, SLOT(addPlayList()));
 
-  connect((QObject *) query, SIGNAL(returnPressed()),
-	  (QObject *) this, SLOT(newQuery()));
-
   connect((QObject *) resultBox, SIGNAL(selected(int)),
 	  (QObject *) this, SLOT(play(int)));
+  connect((QObject *) firstBox, SIGNAL(selectionChanged()),
+	  (QObject *) this, SLOT(firstBoxChanged()));
+  connect((QObject *) secondBox, SIGNAL(selectionChanged()),
+	  (QObject *) this, SLOT(secondBoxChanged()));
+  connect((QObject *) secondBox, SIGNAL(doubleClicked(QListBoxItem *)),
+	  (QObject *) this, SLOT(secondDClicked(QListBoxItem *)));
+  connect((QObject *) firstCombo, SIGNAL(activated(int)),
+	  (QObject *) this, SLOT(firstComboChanged(int)));
+  connect((QObject *) secondCombo, SIGNAL(activated(int)),
+	  (QObject *) this, SLOT(secondComboChanged(int)));
+  connect((QObject *) firstFilterEdit, SIGNAL(textChanged(const QString&)),
+	  (QObject *) this, SLOT(firstFilterChanged(const QString&)));
+  connect((QObject *) firstFilterEdit, SIGNAL(returnPressed()),
+	  (QObject *) this, SLOT(firstBoxChanged()));
+  connect((QObject *) secondFilterEdit, SIGNAL(textChanged(const QString&)),
+	  (QObject *) this, SLOT(secondFilterChanged(const QString&)));
+  connect((QObject *) secondFilterEdit, SIGNAL(returnPressed()),
+	  (QObject *) this, SLOT(secondBoxChanged()));
+  connect((QObject *) resultFilterEdit, SIGNAL(textChanged(const QString&)),
+	  (QObject *) this, SLOT(resultFilterChanged(const QString&)));
+  connect((QObject *) resultFilterEdit, SIGNAL(returnPressed()),
+	  (QObject *) this, SLOT(setPlayList()));
 
+
+  allText=i18n("--All--");
   setCaption(i18n("Database Query"));
+  modeTexts << i18n("Artist") <<i18n("Album") <<i18n("Year") <<i18n("Genre");
+  firstCombo->insertStringList(modeTexts);
+  secondCombo->insertStringList(modeTexts);
+  searchTexts <<"artist" <<"album" <<"year" <<"genre";
 
+  numActiveSearch = 0;
+     
   KWin::setIcons(winId(), *icon, *icon);
+
+  searchThread = new SearchThread(this, db);
+  searchThread->start();
+
+  firstCombo->setCurrentItem((int)ARTIST);
+  secondCombo->setCurrentItem((int)ALBUM);
+  firstComboChanged((int)ARTIST);
+  secondComboChanged((int)ALBUM);
 }
 
 XmmsKdeDBQuery::~XmmsKdeDBQuery() {
 
+  searchThread->kill();
 }
 
 void XmmsKdeDBQuery::setPlayer(PlayerInterface *p) {
@@ -406,54 +464,278 @@ void XmmsKdeDBQuery::setPlayer(PlayerInterface *p) {
   player = p;
 }
 
-void XmmsKdeDBQuery::newQuery() {
-
-  if (!db)
-    return;
-
-  if (!db->isConnectedDB())
-    if (!db->connectDB())
-      return;
-
-  if (searchThread) {
-
-    if (searchThread->finished()) {
-      delete searchThread;
-      searchThread = NULL;
-    }
-  }
-
-  if (!searchThread) {
-    resultBox->clear();
-    searchThread = new SearchThread(db->querydb, this, 
-				    QString(buttons->selected()->name()),
-				    query->text());
-    searchThread->start();
-  }
-}
-
 void XmmsKdeDBQuery::customEvent(QCustomEvent *e) {
 
-  if (e->type() == 60041) {
-    ResultEvent *s = (ResultEvent *) e;
-    QStringList l = s->message();
-    QStringList::Iterator it = l.begin();
-    
-    QString name(*it);
-    it++;
-    QString file(*it);
+  SearchEvent *search = (SearchEvent *)e;
+  QListBoxItem *item;
+  switch (search->type()){
+  case TO_RESULTBOX:
+    resultBox->clear();
+    while (item = search->items.dequeue())
+      resultBox->insertItem(item, 0);
+    break;
 
-    resultBox->insertItem(new QueryItem(name, file));
+  case TO_FIRSTBOX:
+    firstBox->clear();
+    firstBox->insertItem(allText);
+    while (item = search->items.dequeue())
+      firstBox->insertItem(item);
+    break;
+
+  case TO_SECONDBOX:
+    secondBox->clear();
+    secondBox->insertItem(allText);
+    while (item = search->items.dequeue())
+      secondBox->insertItem(item);
+    break;
+
+  case TO_FIRSTBOX_FOCUS:
+    firstBox->clear();
+    firstBox->insertItem(allText);
+    while (item = search->items.dequeue())
+      firstBox->insertItem(item);
+    
+    QListBoxItem *current = firstBox->findItem(search->focusText);
+    firstBox->setCurrentItem(current);
+    firstBox->centerCurrentItem();
+    firstBox->setSelected(current, true);
+    break;
   }
+
+  //kapp->restoreOverrideCursor();
 }
 
+void XmmsKdeDBQuery::activateSearch() {
+   
+  numActiveSearch--;
+
+   if (numActiveSearch == 0) {
+     searchThread->waitCondition.wakeAll();
+     //     kapp->setOverrideCursor(waitCursor);
+   } else
+     searchThread->getSearch();
+}
+
+void XmmsKdeDBQuery::firstFilterChanged(const QString &text) {
+
+   searchThread->addSearch(new SearchEvent(TO_FIRSTBOX, getFirstQuery()), false);
+   numActiveSearch++;
+   if(text.length() > 2)
+     activateSearch();
+   else
+     QTimer::singleShot(400, this, SLOT(activateSearch()));
+}
+
+void XmmsKdeDBQuery::secondFilterChanged(const QString &text) {
+
+   searchThread->addSearch(new SearchEvent(TO_SECONDBOX, getSecondQuery()), false);
+   numActiveSearch++;
+   if(text.length() > 2)
+     activateSearch();
+   else
+     QTimer::singleShot(400, this, SLOT(activateSearch()));
+}
+
+void XmmsKdeDBQuery::resultFilterChanged(const QString &text) {
+
+   searchThread->addSearch(new SearchEvent(TO_RESULTBOX, getResultQuery()), false);
+   numActiveSearch++;
+   if(text.length() > 2)
+     activateSearch();
+   else
+     QTimer::singleShot(400, this, SLOT(activateSearch()));
+}
+
+QString XmmsKdeDBQuery::getFirstQuery() {
+
+  QString lQuery = QString("SELECT DISTINCT %1 FROM music %3 ORDER BY %2 ASC").arg(searchTexts[firstMode]).arg(searchTexts[firstMode]);
+
+if (!firstFilterEdit->text().isEmpty())
+    lQuery = lQuery.arg("WHERE %1 LIKE '%%2%'").arg(searchTexts[firstMode]).arg(firstFilterEdit->text());
+  else
+    lQuery = lQuery.arg("");
+
+  return lQuery;
+}
+
+QString XmmsKdeDBQuery::getSecondQuery() {
+
+  QString lQuery = QString("SELECT DISTINCT %1 FROM music %3 %4 %5 %6 ORDER BY %2 ASC").arg(searchTexts[secondMode]).arg(searchTexts[secondMode]);
+  bool useFirst = true, useFilter = true;
+
+  if (firstList.isEmpty() || firstList.contains(allText))
+    useFirst = false;
+  if (secondFilterEdit->text().isEmpty())
+		useFilter = false;
+
+  if (useFirst || useFilter)
+    lQuery = lQuery.arg("WHERE");
+  else
+    lQuery = lQuery.arg("");
+
+  if (useFirst)
+    lQuery = lQuery.arg("%1 IN %2").arg(searchTexts[firstMode]).arg(prepareList(firstList));
+  else
+    lQuery = lQuery.arg("");
+
+  if (useFirst && useFilter)
+    lQuery = lQuery.arg("AND");
+  else
+    lQuery = lQuery.arg("");
+
+  if (useFilter)
+    lQuery = lQuery.arg("%1 LIKE '%%2%'").arg(searchTexts[secondMode]).arg(secondFilterEdit->text());
+  else
+    lQuery = lQuery.arg("");
+
+  return lQuery;
+}
+
+QString XmmsKdeDBQuery::getResultQuery() {
+
+  QString lQuery = QString("SELECT artist, title, filename FROM music %3 %4 %5 %6 %7 %8 ORDER BY album ASC");
+
+  bool useFirst = true, useSecond = true, useFilter = true;
+
+  if (firstList.isEmpty() || firstList.contains(allText))
+    useFirst = false;
+  if (secondList.isEmpty() || secondList.contains(allText))
+    useSecond = false;
+  if (resultFilterEdit->text().isEmpty())
+    useFilter = false;
+
+  if (useFirst || useSecond || useFilter)
+    lQuery = lQuery.arg("WHERE");
+  else
+    lQuery = lQuery.arg("");
+
+  if (useFirst)
+    lQuery = lQuery.arg("%1 IN %2").arg(searchTexts[firstMode]).arg(prepareList(firstList));
+  else
+    lQuery = lQuery.arg("");
+
+  if (useFirst && useSecond)
+    lQuery = lQuery.arg("AND");
+  else
+    lQuery = lQuery.arg("");
+
+  if (useSecond)
+    lQuery = lQuery.arg("%1 IN %2").arg(searchTexts[secondMode]).arg(prepareList(secondList));
+  else
+    lQuery = lQuery.arg("");
+
+  if ((useFirst || useSecond) && useFilter)
+    lQuery = lQuery.arg("AND");
+  else
+    lQuery = lQuery.arg("");
+
+  if (useFilter)
+    lQuery = lQuery.arg("(artist LIKE '%%1%' OR title LIKE '%%2%')").arg(resultFilterEdit->text()).arg(resultFilterEdit->text());
+  else
+    lQuery = lQuery.arg("");
+
+  return lQuery;
+}
+
+
+
+void XmmsKdeDBQuery::firstBoxChanged() {
+
+  firstList.clear();
+  secondList.clear();	//so getResultQuery wont pull conditions out of this one.
+
+  for (QListBoxItem *lItem = firstBox->firstItem(); lItem; lItem = lItem->next())
+    if (lItem->isSelected())
+      firstList << lItem->text();
+
+  if (!firstFilterEdit->text().isEmpty() && (firstList.empty() || firstList.contains(allText))) {
+
+    firstList.clear(); //rebuild list
+    QListBoxItem *lItem = firstBox->firstItem();
+    while (lItem = lItem->next()) //dont use first item since it's the text ---All---
+      firstList << lItem->text();
+  }
+
+  searchThread->addSearch(new SearchEvent(TO_SECONDBOX, getSecondQuery()));
+  searchThread->addSearch(new SearchEvent(TO_RESULTBOX, getResultQuery()));
+
+  //  kapp->setOverrideCursor(waitCursor);
+}
+
+void XmmsKdeDBQuery::secondBoxChanged() {
+
+  secondList.clear();
+  for (QListBoxItem *lItem = secondBox->firstItem(); lItem; lItem = lItem->next())
+    if (lItem->isSelected())
+      secondList << lItem->text();
+
+  if(!secondFilterEdit->text().isEmpty() && (secondList.empty() || secondList.contains(allText))) {
+
+    secondList.clear(); //rebuild list
+    QListBoxItem *lItem = secondBox->firstItem();
+    while (lItem = lItem->next()) //dont use first item since it's the text ---All---
+      secondList << lItem->text();
+  }
+
+  searchThread->addSearch(new SearchEvent(TO_RESULTBOX, getResultQuery()));
+
+  //  kapp->setOverrideCursor(waitCursor);
+}
+
+void XmmsKdeDBQuery::secondDClicked(QListBoxItem *item) {
+
+  QString text;
+  if (item)
+    text = item->text();
+  else
+    text = allText;
+
+  firstList.clear();
+  secondList.clear();
+
+  firstList << text;
+
+  //swap modes
+  EFilterMode temp = firstMode;
+  firstMode = secondMode;
+  secondMode = temp;
+
+  firstCombo->setCurrentItem((int)firstMode);
+  secondCombo->setCurrentItem((int)secondMode);
+
+  searchThread->addSearch(new SearchEvent(TO_FIRSTBOX_FOCUS, getFirstQuery(), text));
+  searchThread->addSearch(new SearchEvent(TO_SECONDBOX, getSecondQuery()));
+
+  //  kapp->setOverrideCursor(waitCursor);
+
+  firstBox->setFocus();
+}
+
+void XmmsKdeDBQuery::firstComboChanged(int index) {
+
+  firstMode = (EFilterMode) index;
+  searchThread->addSearch(new SearchEvent(TO_FIRSTBOX, getFirstQuery()));
+  //kapp->setOverrideCursor(waitCursor);
+}
+
+void XmmsKdeDBQuery::secondComboChanged(int index) {
+
+  secondMode = (EFilterMode) index;
+  searchThread->addSearch(new SearchEvent(TO_SECONDBOX, getSecondQuery()));
+  //kapp->setOverrideCursor(waitCursor);
+}
 
 void XmmsKdeDBQuery::play(int index) {
 
   if (player) {
-    player->playlistClear();
+
+    if (!click) // set
+      player->playlistClear();
+
     player->playlistAdd(((QueryItem *) resultBox->item(index))->getFile());
-    player->play();
+    
+    if (!click)
+      player->play();
   }
 }
 
@@ -471,14 +753,14 @@ void XmmsKdeDBQuery::setPlayList() {
 
 
 void XmmsKdeDBQuery::addPlayList() {
-  
+
   if (player) {
     QStringList files;
     QStringList allfiles;
-    
+
     for (unsigned int index = 0; index < resultBox->count(); index ++) {
       allfiles += ((QueryItem *) resultBox->item(index))->getFile();
-
+      //printf("   [%s]\n", ((QueryItem *) resultBox->item(index))->getFile().latin1());
       if (resultBox->isSelected(index))
 	files += ((QueryItem *) resultBox->item(index))->getFile();
     }
@@ -492,21 +774,43 @@ void XmmsKdeDBQuery::addPlayList() {
   resultBox->clearSelection();
 }
 
+QString XmmsKdeDBQuery::prepareList(QStringList list) {
+
+  for (unsigned int i = 0; i < list.count(); i++) {
+    QString &string = list[i];
+    for (unsigned int i = 0; i < string.length(); i++) {
+      if (string[i] == '\'') {
+	string.insert(i, '\'');
+	i++;
+      }
+    }
+  }
+  return "(\'" + list.join("\', \'") + "\')";
+}
 
 void XmmsKdeDBQuery::popup() {
 
   if (db->isEnabled()) {
-    int d = KWin::info(winId()).desktop;
-    
-    if (d != KWin::currentDesktop()) {
-      
+
+#if defined(KDE_IS_VERSION)
+#if KDE_IS_VERSION(3, 1, 90)
+    if (!KWin::windowInfo(winId()).isOnCurrentDesktop())
       KWin::setOnDesktop(winId(), KWin::currentDesktop());
-    }
-    
+#endif
+#else
+    if (KWin::info(winId()).desktop != KWin::currentDesktop())
+    KWin::setOnDesktop(winId(), KWin::currentDesktop());
+#endif
     if (!isVisible())
       show();
-    
-    query->setFocus();
+    if (isMinimized())
+      showNormal();
+    if (!isActiveWindow()) {
+      setActiveWindow();
+      raise();
+    }
+
+    firstBox->setFocus();
   }
 }
 
@@ -518,6 +822,7 @@ void XmmsKdeDBQuery::readConfig() {
   frameSize = QSize(200, 320);
   frameSize = config->readSizeEntry("queryframesize", &frameSize);
   pop = config->readNumEntry("popup", 2);
+  click = config->readNumEntry("click", 0);
 
   resize(frameSize);
   move(framePos);
@@ -530,12 +835,13 @@ void XmmsKdeDBQuery::writeConfig() {
   config->writeEntry("queryframeposition", pos());
   config->writeEntry("queryframesize", size());
   config->writeEntry("popup", pop);
+  config->writeEntry("click", click);
 }
 
 QWidget* XmmsKdeDBQuery::getConfigurationWidget(QWidget *parent) {
 
-  QWidget *dbWidget = new QWidget(parent);  
-  
+  QWidget *dbWidget = new QWidget(parent);
+
   QVBoxLayout *vbox = new QVBoxLayout(dbWidget, 10);
 
   queryGroup = new QVButtonGroup(i18n("Show query window"), dbWidget);
@@ -545,8 +851,19 @@ QWidget* XmmsKdeDBQuery::getConfigurationWidget(QWidget *parent) {
   QRadioButton *playerStartup = new QRadioButton(i18n("at player startup"), queryGroup);
   QRadioButton *noStartup = new QRadioButton(i18n("when selected from menu"), queryGroup);
   queryGroup->setButton(pop);
+
   vbox->addWidget(queryGroup);
 
+  clickGroup = new QVButtonGroup(i18n("Doubleclick action"), dbWidget);
+  connect(clickGroup, SIGNAL(clicked(int)), this, SLOT(clickChanged(int)));
+
+  QRadioButton *setB = new QRadioButton(i18n("set as playlist"), clickGroup);
+  QRadioButton *addB = new QRadioButton(i18n("add to playlist"), clickGroup);
+
+  clickGroup->setButton(click);
+
+  vbox->addWidget(clickGroup);
+
   return dbWidget;
 }
 
@@ -555,10 +872,15 @@ void XmmsKdeDBQuery::popupChanged(int index) {
   pop = index;
 }
 
+void XmmsKdeDBQuery::clickChanged(int index) {
+
+  click = index;
+}
+
 
 QueryItem::QueryItem(QString text, QString file):
-
   QListBoxText(text) {
+
   filename = file;
 }
 
@@ -571,69 +893,85 @@ QString QueryItem::getFile() {
   return filename;
 }
 
-SearchThread::SearchThread(sqlite *datab, XmmsKdeDBQuery *q,
-			   QString w, QString l) {
-
-  db = datab;
-  query = q;
-  what = w;
-  like = l;
-}
 
 void SearchThread::run() {
 
+  SearchEvent *search;
+  while (true) {
+
+    //perform all searches that are queued
+    while (search = getSearch())
+      doSearch(search);
+
+    //wait for something to happen
+    waitCondition.wait();
+    if (shouldExit)
+      exit();
+  }
+}
+
+void SearchThread::doSearch(SearchEvent *search) {
+
+  QStringList returnList;
   char *error;
   char **result;
   int nrow;
   int ncolumn;
-  
-  QString text = "%%" + like + "%%";
+  int rc;
+
+  if (!db)
+    return;
+  if (!db->isConnectedDB())
+    if (!db->connectDB())
+      return;
+
+  rc = sqlite_get_table(db->querydb, search->query.latin1(), &result, &nrow, &ncolumn, &error);
+  qDebug("xmms-kde: querying: %s", search->query.latin1());
 
-  int rc = sqlite_get_table_printf(db, "SELECT artist, title, filename FROM music WHERE %q LIKE %Q", &result, &nrow, &ncolumn, &error,
-				   what.latin1(), text.latin1());
-  
   if (error) {
-    qDebug("error: %s\n", error);
+    qDebug("xmms-kde: sqlite error: %s", error);
     free(error);
     error = 0;
   }
 
   if (rc != SQLITE_OK) {
     qDebug("xmms-kde: database query failed");
-  } else {
-    if (nrow >= 1) {
-      for (int index = 1; index <= nrow; index ++) {
+  }
+  else if (nrow >= 1) {
+    qDebug("xmms-kde: num rows in result: %i", nrow);
+
+
+    if ((int)search->type() == (int)TO_RESULTBOX) {
+      for (int i = 1; i <= nrow; i++) {
 	// only add it when the file exists
-	QString fileName(result[index * 3 + 2]);
-
+	QString fileName(result[i * 3 + 2]);
 	if (QFile::exists(fileName)) {
-	  QString name = "[" + QString(result[index * 3]) + "] " +
-	    QString(result[index * 3 + 1]);
-
-	  QStringList l;
-	  l += name;
-	  l += fileName;
-
-	  ResultEvent *event = new ResultEvent(l);
-	  postEvent(query, event);
+	  QString name = QString(result[i*3]) + " - " + QString(result[i*3 + 1]);
+	  search->items.enqueue(new QueryItem(name, QString(result[i*3 +2])));
 	}
       }
-    }        
+    } else {
+      for (int i = 1; i <= nrow; i++)
+	search->items.enqueue(new QListBoxText(QString(result[i])));
+    }
   }
-  
+ 
   sqlite_free_table(result);
+
+  postEvent(receiver, search);
 }
 
 
 InsertThread::InsertThread(sqlite *datab, QStringList dirs,
 			   StatusLabel *status, StatusLabel *file,
-			   ProgressLabel *progess) {
+			   ProgressLabel *progess, QDateTime up) {
   db = datab;
   dir = dirs;
 
   statusLabel = status;
   fileLabel = file;
   progressLabel = progess;
+  updated = up;
 }
 
 void InsertThread::run() {
@@ -657,22 +995,22 @@ void InsertThread::updateDatabase(QStringList dir) {
   t.start();
   r.start();
 
-  // the contents of the database in a hash  
+  // the contents of the database in a hash
   hash_set<const char *, hash<const char *>, eqstr> database;
 
-  // the contents of the filesystem in a hash  
+  // the contents of the filesystem in a hash
   hash_set<const char *, hash<const char *>, eqstr> files;
 
   hash_set<const char*, hash<const char*>, eqstr>::const_iterator it, it2;
 
   QString q("SELECT filename from music");
-  
+
   int nrow, ncolumn;
   char *error;
   char **result;
-  
+
   int rc = sqlite_get_table(db, q.latin1(), &result, &nrow, &ncolumn, &error);
-  
+
   if (error) {
     free(error);
     error = 0;
@@ -682,9 +1020,9 @@ void InsertThread::updateDatabase(QStringList dir) {
   t.start();
 
   database.resize(nrow * 2);
-    
+
   for (int index = 0; index < nrow; index ++) {
-    database.insert(result[index]);    
+    database.insert(result[index]);
   }
 
   printf("xmms-kde: inserted database into hash: %d seconds\n", t.elapsed() /1000);
@@ -694,7 +1032,7 @@ void InsertThread::updateDatabase(QStringList dir) {
   postEvent(statusLabel, event);
 
   vector<QString *> v;
-  
+
   for (QStringList::Iterator it = dir.begin(); it != dir.end(); ++it )
     addPathToVector(*it, &v);
 
@@ -725,7 +1063,7 @@ void InsertThread::updateDatabase(QStringList dir) {
   if (error) {
     free(error);
     error = 0;
-  }  
+  }
 
   int progressIndex = 0;
 
@@ -738,9 +1076,9 @@ void InsertThread::updateDatabase(QStringList dir) {
       ProgressStepEvent *psevent = new ProgressStepEvent(progressIndex);
       postEvent(progressLabel, psevent);
     }
-    
+
     progressIndex ++;
-    
+
     it2 = files.find(*it);
     if (it2 == files.end()) {
       deleteFromDatabase(QString(*it));
@@ -755,7 +1093,7 @@ void InsertThread::updateDatabase(QStringList dir) {
     free(error);
     error = 0;
   }
-  
+
   printf("xmms-kde: removed files from database: %d seconds\n", t.elapsed() / 1000);
   t.start();
 
@@ -775,7 +1113,7 @@ void InsertThread::updateDatabase(QStringList dir) {
     // only generate an event when progressbar should move (>=1% difference)
     if (((float) progressIndex) / size - thresh >= 0.01) {
       thresh = ((float) progressIndex) / size;
-      
+
       ProgressStepEvent *psevent = new ProgressStepEvent(progressIndex);
       postEvent(progressLabel, psevent);
     }
@@ -785,8 +1123,14 @@ void InsertThread::updateDatabase(QStringList dir) {
     it = database.find(v[index]->latin1());
     if (it == database.end()) {
       insertIntoDatabase(*v[index]);
+    } else {
+
+      QFileInfo inf(*v[index]);
+      QDateTime d = inf.lastModified();
+      if (d > updated)
+	insertIntoDatabase(*v[index]);
     }
-  }  
+  }
 
   psevent = new ProgressStepEvent(v.size() + nrow + 1);
   postEvent(progressLabel, psevent);
@@ -799,8 +1143,6 @@ void InsertThread::updateDatabase(QStringList dir) {
     error = 0;
   }
 
-  printf("xmms-kde: inserted into database: %d seconds\n", t.elapsed() / 1000);
-
   event = new StatusEvent(i18n("updating database: done."));
   postEvent(statusLabel, event);
   event = new StatusEvent("");
@@ -817,36 +1159,39 @@ void InsertThread::updateDatabase(QStringList dir) {
 void InsertThread::insertIntoDatabase(QString file) {
 
   QFileInfo inf(file);
-    
+
   if (inf.isFile()) {
 
     KFileMetaInfo m(file);
-	
+
+    if (m.isValid())
+      qDebug("mime: %s\n", m.mimeType().latin1());
+
     if (m.isValid() && (!m.mimeType().endsWith("x-mp3") ||
 			!m.mimeType().endsWith("x-ogg"))) {
-
+      
       QFileInfo f(file);
-      StatusEvent *event = new StatusEvent("[" + m.item("Artist").string() + 
+      StatusEvent *event = new StatusEvent("[" + m.item("Artist").string() +
 					   "] " + m.item("Title").string());
 
       postEvent(fileLabel, event);
-      
-      QString title = m.item("Title").string();
-      QString artist = m.item("Artist").string();
-      QString album = m.item("Album").string();
-      QString track = m.item("Tracknumber").string();
-      QString date = m.item("Date").string();
-      QString genre = m.item("Genre").string();
-      QString comment = m.item("Comment").string();
+
+      QString title = m.item("Title").string().stripWhiteSpace();
+      QString artist = m.item("Artist").string().stripWhiteSpace();
+      QString album = m.item("Album").string().stripWhiteSpace();
+      QString track = m.item("Tracknumber").string().stripWhiteSpace();
+      QString date = m.item("Date").string().stripWhiteSpace();
+      QString genre = m.item("Genre").string().stripWhiteSpace();
+      QString comment = m.item("Comment").string().stripWhiteSpace();
 
       char *error;
-      
-      int rc = sqlite_exec_printf(db, 
+
+      int rc = sqlite_exec_printf(db,
 				  "INSERT INTO music VALUES(%Q, %Q, %Q, %Q, %Q, %Q, %Q, %Q)", 0, 0, &error,
 				  file.latin1(),
-				  title.latin1(), artist.latin1(), 
-				  album.latin1(), track.latin1(), 
-				  date.latin1(), genre.latin1(), 
+				  title.latin1(), artist.latin1(),
+				  album.latin1(), track.latin1(),
+				  date.latin1(), genre.latin1(),
 				  comment.latin1());
 
       if (error) {
@@ -860,34 +1205,169 @@ void InsertThread::insertIntoDatabase(QString file) {
       else
 	printf("xmms-kde: inserted [%s]\n", file.latin1());
       */
-    } else if (file.lower().endsWith(".mp3") ||
-	       file.lower().endsWith(".ogg")) {
-      /* seems to be an ogg or mp3, but doesn't have any meta info 
-	 This could happen if they don't have id3lib installed, or
-	 if there's no id3 tag.   In this case, try to get the
-	 title from the filename. */
-      QString fileName = inf.baseName();
-      fileName = fileName.replace(QRegExp("_"), " ");
-      fileName = fileName.replace(QRegExp("-"), " ");
-      fileName = fileName.stripWhiteSpace();
-      
-      StatusEvent *event = new StatusEvent(fileName);
-      postEvent(fileLabel, event);
-      
+    }
+#ifdef HAVE_ID3LIB
+    else if (file.lower().endsWith("mp3")) {
+      printf("xmms-kde id3lib: %s\n", file.latin1());
+      // use id3lib if available
+      const char *artist = NULL;
+      const char *title = NULL;
+      const char *album = NULL;
+      const char *track = NULL;
+      const char *date = NULL;
+
+      ID3_Tag tag(file.latin1());
+      ID3_Frame *frame = NULL;
+      ID3_Field *field = NULL;
+
+      frame = tag.Find(ID3FID_LEADARTIST);
+      if (frame) {
+	field = frame->GetField(ID3FN_TEXT);
+	if (field) {
+	  artist = field->GetRawText();
+	}
+      }
+
+      frame = tag.Find(ID3FID_TITLE);
+      if (frame) {
+	field = frame->GetField(ID3FN_TEXT);
+	if (field) {
+	  title = field->GetRawText();
+	}
+      }
+
+      frame = tag.Find(ID3FID_ALBUM);
+      if (frame) {
+	field = frame->GetField(ID3FN_TEXT);
+	if (field) {
+	  album = field->GetRawText();
+	}
+      }
+
+      frame = tag.Find(ID3FID_TRACKNUM);
+      if (frame) {
+	field = frame->GetField(ID3FN_TEXT);
+	if (field) {
+	  track = field->GetRawText();
+	}
+      }
+
+      frame = tag.Find(ID3FID_DATE);
+      if (frame) {
+	field = frame->GetField(ID3FN_TEXT);
+	if (field) {
+	  date = field->GetRawText();
+	}
+      }
+
       char *error;
-      
-      int rc = sqlite_exec_printf(db, 
+
+      int rc = sqlite_exec_printf(db,
 				  "INSERT INTO music VALUES(%Q, %Q, %Q, %Q, %Q, %Q, %Q, %Q)", 0, 0, &error,
-				  file.latin1(), fileName.latin1(), "", "", "", "", "", "");
+				  file.latin1(),
+				  title, artist, album, track, date, "", "");
 
       if (error) {
 	free(error);
 	error = 0;
-      }     
+      }
+
+      StatusEvent *event = new StatusEvent("[" + QString(artist) +
+					   "] " + QString(title));
+      postEvent(fileLabel, event);
+
+      /*
+      delete title;
+      delete artist;
+      delete album;
+      delete track;
+      delete date;
+      */
 
       if (rc != SQLITE_OK)
  	qDebug("xmms-kde: database insert failed on [%s]\n", file.latin1());
     }
+#endif
+#ifdef HAVE_VORBIS
+    else if (file.lower().endsWith("ogg")) {
+      printf("xmms-kde vorbis: %s\n", file.latin1());
+      FILE *fh;
+
+      if ((fh = fopen(file.latin1(), "r")) != NULL) {
+	OggVorbis_File *vf = (OggVorbis_File *) malloc(sizeof(OggVorbis_File));
+	if (ov_open(fh, vf, NULL, 0) == 0) {
+
+	  vorbis_comment *comment = 0;
+	  comment = ov_comment(vf, -1);
+
+	  char **ptr = comment->user_comments;
+
+	  char *title_utf8 = 0;
+	  char *artist_utf8 = 0;
+	  char *album_utf8 = 0;
+	  char *track_utf8 = 0;
+	  char *date_utf8 = 0;
+	  char *genre_utf8 = 0;
+	
+	  for (int i = 0; i < comment->comments; i++, ptr++) {
+
+	    if (!*ptr)
+	      continue;
+	    
+	    if (!strncasecmp(*ptr, "ARTIST=", 7))
+	      artist_utf8 = (*ptr) + 7;
+
+	    else if (!strncasecmp(*ptr, "TITLE=", 6))
+	      title_utf8 = (*ptr) + 6;
+
+	    else if (!strncasecmp(*ptr, "ALBUM=", 6))
+	      album_utf8 = (*ptr) + 6;
+
+	    else if (!strncasecmp(*ptr, "DATE=", 5))
+	      date_utf8 = (*ptr) + 5;
+
+	    else if (!strncasecmp(*ptr, "GENRE=", 6))
+	      genre_utf8 = (*ptr) + 6;
+
+	    else if (!strncasecmp(*ptr, "TRACKNUMBER=", 12))
+	      track_utf8 = (*ptr) + 12;
+	  }
+
+	  QString title = QString::fromUtf8(title_utf8);
+	  QString artist = QString::fromUtf8(artist_utf8);
+	  QString album = QString::fromUtf8(album_utf8);
+	  QString track = QString::fromUtf8(track_utf8);
+	  QString date = QString::fromUtf8(date_utf8);
+	  QString genre = QString::fromUtf8(genre_utf8);
+	
+	  char *error;
+	  
+	  int rc = sqlite_exec_printf(db,
+				      "INSERT INTO music VALUES(%Q, %Q, %Q, %Q, %Q, %Q, %Q, %Q)", 0, 0, &error,
+				      file.latin1(),
+				      title.latin1(), artist.latin1(), album.latin1(), track.latin1(), date.latin1(), genre.latin1(), "");
+	  
+	  if (error) {
+	    free(error);
+	    error = 0;
+	  }
+
+	  StatusEvent *event = new StatusEvent("[" + QString(artist) +
+					       "] " + QString(title));
+	  postEvent(fileLabel, event);
+
+	  if (rc != SQLITE_OK)
+	    qDebug("xmms-kde: database insert failed on [%s]\n", file.latin1());
+	} else {
+	  fclose(fh);
+	}
+	
+	if (vf) {
+	  ov_clear(vf);
+	}
+      }
+    }
+#endif
   }
 }
 
@@ -895,9 +1375,9 @@ void InsertThread::insertIntoDatabase(QString file) {
 void InsertThread::deleteFromDatabase(QString file) {
 
   char *error;
-      
+
   int rc = sqlite_exec_printf(db, "DELETE FROM music WHERE filename = %Q", NULL, 0, &error, file.latin1());
-      
+
   if (error) {
     free(error);
     error = 0;
@@ -916,15 +1396,15 @@ void InsertThread::addPathToVector(QString path, vector<QString *> *v) {
   QDir dir(path);
   if (!dir.exists())
     return;
-      
+
   QStringList fileList(dir.entryList());
-  
-  for (QStringList::Iterator it = fileList.begin(); it != fileList.end(); ++it ) {
+
+  for (QStringList::Iterator it = fileList.begin(); it != fileList.end(); ++it) {
     QString absfile(dir.absPath() + "/" + (*it));
     QString file(*it);
-    
+
     QFileInfo inf(absfile);
-    
+
     if (inf.isDir()) {
       if (!file.startsWith(".")) {
 	addPathToVector(absfile, v);
@@ -932,10 +1412,9 @@ void InsertThread::addPathToVector(QString path, vector<QString *> *v) {
     } else if (inf.isFile()
 	       && ((file.endsWith(".mp3") || file.endsWith(".ogg") ||
 		    file.endsWith(".MP3") || file.endsWith(".OGG")))) {
-      
+
       v->push_back(new QString(absfile));
     }
   }
 }
 
-
